/* 
    Copyright (C) 2006 by Gary A. Morris
    
    This file is part of CAN_decoder - http://mggm.net/prius/candecoder.html

    CAN_decoder is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CAN_decoder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CAN_decoder; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <exception>
#include <cstdio>

#include "globals.h"
#include "CANDataMgr.h"


CANDataMgr::CANDataMgr(const Globals::UnitsType units) : headerPrinted(false)
{
    CANFields[timestamp          ].init( "Time" );
    // Note: range limits are only used to determine if a packet is corrupted
    //       so they should allow for all possible values
    CANFields[voltage            ].init( "Voltage",                100,   400 );
    CANFields[soc                ].init( "SOC %",                    0,   100 );
    CANFields[amps               ].init( "Amps",                  -200,   200 );
    CANFields[maxChargeCurrent   ].init( "Max Charge Current",       0,   200 );
    CANFields[maxDischargeCurrent].init( "Max Discharge Current", -200,     0 );
    CANFields[minBattTemp        ].init( "Min Batt Temp C",          0,   250 );
    CANFields[maxBattTemp        ].init( "Max Batt Temp C",          0,   250 ); // may be C or F
    CANFields[speed              ].init( "Speed (km/h)",          -100,   250 );
    CANFields[speedometer        ].init( "Speedometer",           -100,   150 );
    CANFields[speedRelated       ].init( "Speed Related",       -10000, 32767 );
    CANFields[frontLeft          ].init( "FL Wheel",            -10000, 32767 );
    CANFields[frontRight         ].init( "FR Wheel",            -10000, 32767 );
    CANFields[rearLeft           ].init( "RL Wheel",            -10000, 32767 );
    CANFields[rearRight          ].init( "RR Wheel",            -10000, 32767 );
    CANFields[iceRPM             ].init( "ICE RPM",                  0,  8000 );
    CANFields[iceTemp            ].init( "ICE Temp C",               0,   299 );
    CANFields[waterTemp          ].init( "Coolant Temp C",           0,   299 );
    CANFields[sideAccel          ].init( "Side Accel",            -512,   512 );
    CANFields[sideAccel2         ].init( "Side Accel1",           -512,   512 );
    CANFields[fwdAccel           ].init( "Fwd Accel",             -512,   512 );
    CANFields[fwdAccel2          ].init( "Fwd Accel2",            -512,   512 );
    CANFields[steering           ].init( "Steering",             -1000,  1000 );
    CANFields[gasPedal           ].init( "Gas Pedal",                0,   255 );
    CANFields[brakePedal         ].init( "Brake Pedal",              0,   127 );
    CANFields[brakesPressed      ].init( "Brake Pressed",            0,     1 );
    CANFields[shiftLever         ].init( "Shifter",                 -2,     2 );
    CANFields[evMode             ].init( "EV Mode",                  0,     0 );
    CANFields[fuelInj            ].init( "Fuel Inj",                 0, 32767 );
    CANFields[fuelFlow           ].init( "Fuel Flow (L/hr)",         0,   255 );
    CANFields[VVTi               ].init( "VVTi",                  -255,   255 );
    CANFields[gasGauge           ].init( "Gas Gauge",                0,   100 );
    CANFields[cruiseControl      ].init( "Cruise Control",           0,     1 );
    CANFields[lights             ].init( "Lights",                   0,     3 );
	CANFields[kmlMPGnum			 ].init( "iKM/L",				 -9999,   9999);
    
    if (units == Globals::US) {
        CANFields[speed      ].setName("Speed (mph)"     );
		CANFields[fuelFlow   ].setName("Fuel Flow (G/hr)");
        CANFields[minBattTemp].setName("Min Batt Temp F" );
        CANFields[maxBattTemp].setName("Max Batt Temp F" );
        CANFields[iceTemp    ].setName("ICE Temp F"      );
        CANFields[waterTemp  ].setName("Coolant Temp F"  );
		CANFields[kmlMPGnum  ].setName("iMPG"			 );
    }
    
    // check to make sure we have set the fieldName for all fields
    for (int i=0; i<NFIELDS; i++) {
        if (!CANFields[i].isInitialized()) {
            std::fprintf(stderr, "*** ERROR: CANDataMgr: uninitialized field entry #%d\n", i );
        }
    }
}

void CANDataMgr::updateField(FieldIdType fieldId, double fieldValue) {
    if (CANFields[fieldId].isInRange(fieldValue)) {
        char fieldString[80];
        std::sprintf(fieldString, "%1g", fieldValue);
    
        updateField(fieldId, fieldString, false);
        
    } else {
        fprintf(stderr, "*** WARNING: bad data, field out of range, %s = %g\n",
                CANFields[fieldId].name().c_str(),
                fieldValue);
    }
}
        

void CANDataMgr::updateField(FieldIdType fieldId, const std::string& fieldValue, bool addQuotes) {
    if (fieldId >= NFIELDS) throw std::exception();
    
    if (Globals::debug & 0x02) {
        fprintf(stderr, "--- %s = %s\n", CANFields[fieldId].name().c_str(), fieldValue.c_str());
    }
    
    if (addQuotes) {
        CANFields[fieldId].value( "\"" + fieldValue + "\"" );
    } else {
        CANFields[fieldId].value( fieldValue );
    }
}

void CANDataMgr::clearFields() {
    for (int i=0; i<NFIELDS; i++) {
        CANFields[i].clearValue();
    }
}


void CANDataMgr::outputFields(std::FILE* outFile) {
    if (outFile == NULL) throw std::exception();
    
    static long timeStamp = 0;
    updateField(timestamp, ++timeStamp);
    
    if (!headerPrinted) {
        for (int i=0; i<NFIELDS; i++) {
            if (i > 0) {
                fprintf(outFile, ",");
            }
            fprintf(outFile, "\"%s\"", CANFields[i].name().c_str());
        } 
        
        fprintf(outFile, "\n");
        headerPrinted = true;
    }
    
    // check and see if we have a valid set of data, if too many fields are not set or invalid
    // then skip this set of data
    int nValid   = 0;
    int nInvalid = 0;
    
    for (int i=1; i<NFIELDS; i++) {
        if (CANFields[i].isValid()) {
            nValid++;
        } else {
            nInvalid++;
        }
    }
    
    if (nValid >= 8 || nInvalid <= nValid) {
        if (Globals::verbose) {
            fprintf(stderr, "--- valid packet %d -------------------------\n", timeStamp);
        }
        
    } else {
        fprintf(stderr, 
                "*** WARNING: skipping data packet, %d fields invalid, %d fields valid\n",
                nInvalid, nValid );
        return;  // too many invalid data fields
    }
    
    for (int i=0; i<NFIELDS; i++) {
        if (i > 0) {
            fprintf(outFile, ",");
        }
        if (CANFields[i].isValid()) {
            fprintf(outFile, "%s", CANFields[i].value().c_str());
        }
    } 
    
    fprintf(outFile, "\n");
}



    
    