/* 
    Copyright (C) 2006 by Gary A. Morris
    
    This file is part of CAN_decoder - http://mggm.net/prius/candecoder.html

    CAN_decoder is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CAN_decoder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CAN_decoder; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	Modified by Jay Groh to work with the V4 CAN-View. http://www.jaygroh.com
*/


//--------------------------------------------------------------
//  To recognize new packets, modify:
//
//    (1) CAN_packets.cpp - add id/length to headers array
//
//--------------------------------------------------------------
//  To add a new field display, modify:
//
//    (1) CANDataMgr.h    - FieldIndexType  <-- add field id
//    (2) CANDataMgr.cpp  - CANDataMgr()    <-- init field name
//    (3) CAN_packets.cpp - decodePacket()  <-- call updateField
//  
//--------------------------------------------------------------



#include <cstdio>
#include <cerrno>
#include <string>
#include <cctype>
#include <iostream>
#include <map>


#include "globals.h"
#include "FieldData.h"
#include "CAN_packets.h"
#include "CANDataMgr.h"
const char* const version = "V4 1.8.2";

CANPackets packets; 
CANDataMgr* dataMgr = NULL; 

std::FILE* inFile;
std::FILE* outFile;

//------------------------------------------------------------------------
//
// fillBuffer - reads data from the input file to fill a packet buffer
//
// returns SUCCESS if successful, 
// returns FAILED if there is nothing to read or failure
//------------------------------------------------------------------------
int fillBuffer(CANPackets::BufferType* bufP,  // points to beginning of buffer
               unsigned bufSize) {            // size of the buffer area
    
    unsigned bytesRead = fread(bufP, 1, bufSize, inFile);
    
    if (bytesRead != bufSize) {
        if (ferror(inFile) || feof(inFile)) {
            return rFAILED;
        }
    }
    return rSUCCESS;
}

//------------------------------------------------------------------------
//
// shiftAndFill - shifts bytes within the packet buffer
//
//------------------------------------------------------------------------
int shiftAndFill(CANPackets::BufferType* buffer) {
    // shift the bytes by one and try again
    memcpy( &buffer[0], &buffer[1], HEADERSIZE-1);
    
    // read another byte to complete the header
    return fillBuffer(&buffer[2], 1);
}

//------------------------------------------------------------------------
//
// readData - reads packets from the input file and decodes them
//
//------------------------------------------------------------------------
void readData() {
    CANPackets::BufferType buffer[PACKETSIZE];
    PacketIdType packetId = 0;
    long skipped = 0;   // keep track of how many bytes we skip
    bool done = false;
    bool needStartPacket = true;
    int  lowestPacketId  = MAX_PACKET_ID;
    int  highestPacketId = MIN_PACKET_ID;
    
    // read the packet header only
    int result = fillBuffer(buffer, HEADERSIZE);    
    if (result != rSUCCESS) {
        fprintf(stderr,"*** ERROR: unable to read file");
        return;
    }
        
    // decode all the packets
    while (!done) {
        packetId = packets.isPacket(buffer);
        
        if (needStartPacket && packetId != START_PACKET_ID) {
            // shift the bytes by one, and see if we can find the start of
            // a packet header
            if (shiftAndFill(buffer) == rFAILED) {
                done = true;
                break;
            } else {
                skipped++;
            }
        
        } else if (!needStartPacket && packetId == rFAILED) {
            // we're in the middle of a burst of packet but we don't
            // recognize this packet, skip over it the bad packet's data
            result = fillBuffer(&buffer[HEADERSIZE], PACKETSIZE-HEADERSIZE);
            if (result != rSUCCESS) {
                done = true;
                break;
            }
            skipped += PACKETSIZE;
            
            // start reading the next packet header
            result = fillBuffer(buffer, HEADERSIZE);    
            if (result != rSUCCESS) {
                done = true;
                break;
            }                        

        } else {
            // we have a valid packet
            needStartPacket = false;
            
            // if we skipped any, tell how many
            if (skipped) {
                if (Globals::verbose) {
                    fprintf(stderr, "*** skipped %d bytes\n", skipped);
                }
                
                if (skipped >= PACKETSIZE * 80) {
                    // we skipped more than a whole packet, dump what we have an start
                    // fresh
                    fprintf(stderr, "*** WARNING: skipped bad packet, writing out unfinished packet\n");
                    dataMgr->outputFields(outFile);
                    dataMgr->clearFields();
                    
                    // reset limits for the new packet
                    lowestPacketId  = packetId;
                    highestPacketId = packetId;
                }
                
                skipped = 0;
            }
            
            // we only have the header, read the rest of the packet
            result = fillBuffer(&buffer[HEADERSIZE], PACKETSIZE-HEADERSIZE);
            if (result != rSUCCESS) {
                done = true;
                break;
            }
            
            if (packetId >= 0x20 && packetId <= 0x3cd) {
                // if this looks like the start of a newer packet, output the 
                // previously collected data
                if (lowestPacketId < highestPacketId && packetId <= lowestPacketId) {
                    fprintf(stderr, "*** WARNING: writing out unfinished packet\n");
                    dataMgr->outputFields(outFile);
                    dataMgr->clearFields();
                    
                    // reset limits for the new packet
                    lowestPacketId  = packetId;
                    highestPacketId = packetId;
                }
            
                // record hi and lo packet ids
                if (packetId < lowestPacketId ) lowestPacketId  = packetId;
                if (packetId > highestPacketId) highestPacketId = packetId;
            }
            
            // decode the packet data
            packets.decodePacket(buffer, packetId, *dataMgr); 
            
            // if this is the last packet of a burst, output the data
            if (packetId == END_PACKET_ID) {
                if (Globals::verbose) {
                    fprintf(stderr, "--- writing out finished packet\n");
                }
                dataMgr->outputFields(outFile);
                dataMgr->clearFields();
                
                // reset limits and need start packet flag
                lowestPacketId  = MAX_PACKET_ID;
                highestPacketId = MIN_PACKET_ID;
                needStartPacket = true;
            }
            
            // start reading a new packet
            result = fillBuffer(buffer, HEADERSIZE);    
            if (result != rSUCCESS) {
                done = true;
                break;
            }
        }
    }
}

void show_command_syntax() {
    fprintf(stderr, "command line options: \n");
    fprintf(stderr, "   [-v] [-d] [-D] [-US] [<input file>] [<output file>]\n\n");
    fprintf(stderr, "   -v  verbose output\n" );
    fprintf(stderr, "   -d  debug output (hex dump of CAN packet contents)\n");
    fprintf(stderr, "   -D  extra debug output (print out CAN field values as decoded)\n");
    fprintf(stderr, "   -US convert values to US units (miles and degrees F)\n\n");
    fprintf(stderr, "   <input file> specifies the name of the input binary data file,\n");
    fprintf(stderr, "                optional, defaults to test_raw.bin if not specified.\n\n");
    fprintf(stderr, "   <outputfile> specifies the name of the output csv file,\n");    
    fprintf(stderr, "                optional, defaults to input name with .csv suffix.\n");
}

int main(int argc, char** argv)
{  
    std::string inFileName  = "";
    std::string outFileName = "";
    
    fprintf(stderr, "Prius CAN-View Raw Format Data Converter\n");
    fprintf(stderr, "version %s  (build %s)\n", version, __DATE__ );
    fprintf(stderr, "(C) Copyright 2006 Gary A. Morris - www.mggm.net\n");
	fprintf(stderr, "Modified by Jay Groh for the V4 CAN-View - www.jaygroh.com\n\n");					// Added my info to let you know this is a modified version
       
    for (int arg=1; arg<argc; arg++) {
        if (strcmp(argv[arg],"-d")==0) {
            Globals::debug |= 0x01;
            
        } else if (strcmp(argv[arg],"-D")==0) {
            Globals::debug |= 0x02;
            
        } else if (strcmp(argv[arg],"-v")==0) {
            Globals::verbose = true;
        
        } else if (strcmp(argv[arg],"-US")==0) {
            Globals::units = Globals::US;
        
        } else if (strcmp(argv[arg],"-UK")==0) {
            Globals::units = Globals::Imperial;
        
        } else if (argv[arg][0] != '-') {
            // (U) arg doesn't start with a hyphen, so assume its a filename
            if (inFileName.empty()) {
                inFileName = argv[arg];
                
            } else if (outFileName.empty()) {
                outFileName = argv[arg];
                
            } else {
                std::fprintf(stderr, "*** FAIL: bad argument: %s\n", argv[arg]);
                show_command_syntax();
                return -1;
            }
        } else {
            show_command_syntax();
            return -1;
        }
    }             
    
    // (U) Create the CANDataMgr object now that we know what units we want
    dataMgr = new CANDataMgr(Globals::units);    
    
    // (U) If no input filename was specified, prompt the user to input a file name
    if (inFileName.empty()) {
        std::cout << "Enter input filename: " << std::flush;
        
        std::getline(std::cin, inFileName);
        
        if (inFileName.empty()) {
            inFileName = "test_raw.bin";
        }
    }
    
    // (U) If no output filename was specified use the input filename but change the suffix
    if (outFileName.empty()) {
        // (U) Remove the suffix from the input filename and put ".csv" suffix instead
        const unsigned len = inFileName.length();
        const unsigned suffix = inFileName.rfind(".",len);
        
        if (len>4 && suffix<len) {
            // (U) Replace the suffix
            outFileName = inFileName.substr(0, suffix) + ".csv";
            
        } else {            
            outFileName = "test_raw.csv";
        }                
    }    
    
    //-----------------------------------------
    // Open the input data file
    //-----------------------------------------
    inFile = std::fopen(inFileName.c_str(),"rb");
    
    if (inFile == NULL) {
        const std::string errMsg = "*** FAIL: opening input file " + inFileName;
        std::perror(errMsg.c_str());
        return -1;
       
    } else {
        const std::string msg = "--- opened input file " + inFileName;
        std::fprintf(stderr, "%s\n", msg.c_str());
    }
	
    //-----------------------------------------
    // Open the output data file
    //-----------------------------------------
    outFile = std::fopen(outFileName.c_str(),"w");
    
    if (outFile == NULL) {
        const std::string errMsg = "*** FAIL: opening output file " + outFileName;
        std::perror(errMsg.c_str());
        return -1;
       
    } else {
        std::fprintf(stderr, "--- opened out file %s\n\n", outFileName.c_str());
    }
	
	//-----------------------------------------
	// Find the start of a packet
	//-----------------------------------------
	readData();
	
    //-----------------------------------------
    // Close the data files
    //-----------------------------------------
	std::fclose( inFile  );
	std::fclose( outFile );
	return 0;
}
