/* 
    Copyright (C) 2006 by Gary A. Morris
    
    This file is part of CAN_decoder - http://mggm.net/prius/candecoder.html

    CAN_decoder is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CAN_decoder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CAN_decoder; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	Modified by Jay Groh to work with the V4 CAN-View. http://www.jaygroh.com
*/

#include "globals.h"
#include "FieldData.h"
#include "CAN_packets.h"
#include "CANDataMgr.h"
#include <round.h>					// Needed in Visual Studio to round numbers.
#include <cstdio>
#include <cmath>

double fuelFlowNum, kmlMPG, speedNum;	// For km/l and MPG calculation


typedef struct {
    PacketIdType  id;   // packet id
    PacketLenType len;  // length of the data
} HeaderDataType;

const HeaderDataType 
            headers[] = { // id    len
                          { 0x0002, 4 },
                          { 0x0020, 3 },
                          { 0x0022, 8 },
                          { 0x0023, 7 },
                          { 0x0025, 8 },
                          { 0x0030, 8 },
                          { 0x0038, 7 },
                          { 0x0039, 4 },
                          { 0x003a, 7 },
                          { 0x003b, 5 },
                          { 0x003e, 3 },
                          { 0x0060, 7 },
                          { 0x0087, 8 },
                          { 0x00b0, 3 },
                          { 0x00b1, 6 },
                          { 0x00b3, 6 },
                          { 0x00b4, 8 },
                          { 0x00c9, 5 },
                          { 0x0120, 8 },
                          { 0x0230, 7 },
                          { 0x0244, 8 },
                          { 0x0262, 4 },
                          { 0x0348, 6 },
                          { 0x034f, 5 },
                          { 0x03c8, 5 },
                          { 0x03c9, 8 },
                          { 0x03ca, 5 },
                          { 0x03cb, 7 },
                          { 0x03cd, 5 },
                          { 0x03cf, 5 },
                          { 0x0423, 1 },
                          { 0x04c1, 8 },
                          { 0x04c3, 8 },
                          { 0x04c6, 8 },
                          { 0x04c7, 8 },
                          { 0x04c8, 8 },
                          { 0x04ce, 8 },
                          { 0x04d0, 8 },
                          { 0x04d1, 8 },
                          { 0x0520, 3 },
                          { 0x0521, 2 },
                          { 0x0526, 3 },
                          { 0x0527, 3 },
                          { 0x0528, 4 },
                          { 0x0529, 7 }, //changed from 6 to 7 for V4 CAN-View
                          { 0x052c, 2 },
                          { 0x0540, 4 },
                          { 0x0553, 7 },
                          { 0x0554, 7 },
                          { 0x056d, 4 },
                          { 0x057f, 7 },
                          { 0x0591, 4 },
                          { 0x05a4, 2 },
                          { 0x05b2, 4 },
                          { 0x05b6, 3 },
                          { 0x05c8, 3 },
                          { 0x05cc, 3 },
                          { 0x05d4, 2 },
                          { 0x05ec, 7 },
                          { 0x05ed, 6 },
                          { 0x05f8, 2 },
                          { 0x0602, 2 }
                        };
        
const int NHEADERS = sizeof(headers) / sizeof(HeaderDataType);


//------------------------------------------------------
// Extract a 12 bit signed value from the buffer,
// 1st byte is the MSB, 2nd byte is LSB
//------------------------------------------------------
Signed16Type CANPackets::getSigned12(const BufferType* bytes) const {
                        
    Signed16Type value = Signed16Type(bytes[0] << 8 | bytes[1]);
    
    if (value & 0x0800) {
        // sign bit (b11) is set, sign extend to the long's sign bit (b31)
        value |= 0xF000;
    } 
    
    return value;
}
        
//------------------------------------------------------
// Extract a 16 bit unsigned value from the buffer,
// 1st byte is the MSB, 2nd byte is LSB
//------------------------------------------------------
Unsigned16Type CANPackets::getUnsigned16(const BufferType* bytes) const {
    return Unsigned16Type(bytes[0] << 8 | bytes[1]);
}

//------------------------------------------------------
// Extract a 16 bit signed value from the buffer,
// 1st byte is the MSB, 2nd byte is LSB
//------------------------------------------------------
Signed16Type CANPackets::getSigned16(const BufferType* bytes) const {
                        
    return Signed16Type( getUnsigned16(bytes) );
}

//------------------------------------------------------
// Convert km/h to MPH 
//------------------------------------------------------
double CANPackets::convertSpeed(const double speedKMH) const {
    switch (Globals::units) {
        case Globals::US:
            return speedKMH * 0.621371;   // convert km/hr to miles/hr

        case Globals::Imperial:
        case Globals::Metric:
        default:  
            return speedKMH;
    }    
}

//------------------------------------------------------
// Convert Liters to Gallons
//------------------------------------------------------
double CANPackets::convertFuel(const double fuelLiters) const {
    switch (Globals::units) {
        case Globals::US:
            return fuelLiters * 0.264172052;   // convert Liters to Gallons

        case Globals::Imperial:
        case Globals::Metric:
        default:  
            return fuelLiters;
    }    
}

//------------------------------------------------------
// convert temp in degrees C into desired units based
// on global units flag
//------------------------------------------------------
long CANPackets::convertTemp(const double degreesC) const {
    switch (Globals::units) {
        case Globals::US:
//            return std::lround((degreesC + 17.7778) * 1.8);		// Removed to compile in VS
			  return Round((degreesC + 17.7778) * 1.8);				// Replacement code to compile in VS
            
        case Globals::Imperial:
        case Globals::Metric:
        default:  
//            return std::lround(degreesC);							// Removed to compile in VS
			  return Round(degreesC);								// Replacement code to compile in VS
    }
}        

//------------------------------------------------------
// Prints out the packet bytes for debugging
//------------------------------------------------------
void CANPackets::printPacket(const PacketIdType id,
                             const PacketLenType len,
                             const BufferType* bytes) const {
                             
    fprintf(stderr, "%04X:", id);
    
    for (int i=0; i<len; i++) {
        fprintf(stderr, " %02X", bytes[i]);
    }
    
    fprintf(stderr, "\n");    
}
        
//-----------------------------------------------------------------
// decodePacket extracts the content of the packet to the
// output file in CSV format
//-----------------------------------------------------------------
void CANPackets::decodePacket(const BufferType* packetP,
                              const PacketIdType packetId,
                              CANDataMgr& data) const {
    
    const BufferType* bytes = &packetP[HEADERSIZE];  // point past header
    
    if (Globals::debug) {
        printPacket( packetId, packetP[2], bytes );
    }
        
    //---------------------------------------
    // 22 - Side Accel
    //---------------------------------------
    if (packetId == 0x22) {
        // subtract 0x200 so that 0 accel is 0, 
        // right accel is positive, 
        // left is negative
        data.updateField(CANDataMgr::sideAccel,  getUnsigned16(&bytes[0]) - 0x200);
        data.updateField(CANDataMgr::sideAccel2, getUnsigned16(&bytes[2]) - 0x200);

    //---------------------------------------
    // 23 - Forward Accel
    //---------------------------------------
    } else if (packetId == 0x23) {
        // subtract 0x200 so that 0 accel is 0, 
        // forward accel is positive, 
        // decel is negative
        data.updateField(CANDataMgr::fwdAccel,  getUnsigned16(&bytes[0]) - 0x200);
        data.updateField(CANDataMgr::fwdAccel2, getUnsigned16(&bytes[2]) - 0x200);

    //---------------------------------------
    // 25 - Steering Wheel Position
    //---------------------------------------
    } else if (packetId == 0x25) {
        int steeringCenter = getSigned12(&bytes[2]);  // value at center position
        int steeringValue  = getSigned12(&bytes[0]) - steeringCenter;
        
        data.updateField(CANDataMgr::steering, steeringValue);

    //---------------------------------------
    // 30 - Brake Pedal
    //---------------------------------------
    } else if (packetId == 0x30) {
        int brakePedal = bytes[4];
        
        data.updateField(CANDataMgr::brakePedal, Round(brakePedal / 1.269));					// Devided by 1.269 to get 0-100 percentage
        
        int pressed = (bytes[0] & 0x80) ? 0 : 1;
        
        data.updateField(CANDataMgr::brakesPressed, pressed);

    //---------------------------------------
    // 38 - Fuel Flow
    //---------------------------------------
    } else if (packetId == 0x38) {
		
		fuelFlowNum = Round(convertFuel(bytes[1]) / 10, .01);									// setting fuelFlowNum

		data.updateField(CANDataMgr::fuelFlow, fuelFlowNum);			// Get L/hr by dividing by 10 then round it to the nearist .01

    //---------------------------------------
    // 39 - ICE Temp
    //---------------------------------------
    } else if (packetId == 0x39) {
        data.updateField(CANDataMgr::iceTemp, convertTemp(bytes[0]));

    //---------------------------------------
    // 3B - Current/Voltage
    //---------------------------------------
    } else if (packetId == 0x3b) {
        long amps = -getSigned12(bytes);
        
        data.updateField(CANDataMgr::amps, amps * 0.1);
        data.updateField(CANDataMgr::voltage, getUnsigned16(&bytes[2]));
    
    //---------------------------------------
    // B1 - Front Wheel Speeds
    //---------------------------------------
    } else if (packetId == 0xb1) {
        // values are speed in kph * 100
        data.updateField(CANDataMgr::frontLeft,  convertSpeed(getSigned16(&bytes[2]) / 100.0));
        data.updateField(CANDataMgr::frontRight, convertSpeed(getSigned16(&bytes[0]) / 100.0));
    
    //---------------------------------------
    // B3 - Rear Wheel Speeds
    //---------------------------------------
    } else if (packetId == 0xb3) {
        // values are speed in kph * 100
        data.updateField(CANDataMgr::rearLeft,  convertSpeed(getSigned16(&bytes[2]) / 100.0));
        data.updateField(CANDataMgr::rearRight, convertSpeed(getSigned16(&bytes[0]) / 100.0));
    
    //---------------------------------------
    // B4 - Speed
    //---------------------------------------
    } else if (packetId == 0xb4) {
        long   speedL = getSigned16(&bytes[5]);
        double speedD = (speedL / double(0x400)) * 10.0;
        
		speedNum = convertSpeed(speedD);

		data.updateField(CANDataMgr::speed, speedNum );
    
    //---------------------------------------
    // 244 - Gas Pedal
    //---------------------------------------
    } else if (packetId == 0x244) {
        
        data.updateField(CANDataMgr::gasPedal, bytes[6]);
        data.updateField(CANDataMgr::speedRelated, getSigned16(&bytes[4]));
        
    //---------------------------------------
    // 3C8 - ICE RPM
    //---------------------------------------
    } else if (packetId == 0x3c8) {
        long iceRPM = bytes[3] << 8 | bytes[2];
        
        data.updateField(CANDataMgr::iceRPM, iceRPM * 32);
    
    //---------------------------------------
    // 3CA - Speedometer
    //---------------------------------------
    } else if (packetId == 0x3ca) {
        
//        data.updateField(CANDataMgr::speedometer, std::lround(convertSpeed(bytes[2])));	// removed to compile in VS
		  data.updateField(CANDataMgr::speedometer, Round(convertSpeed(bytes[2])));

    //---------------------------------------
    // 3CB - Battery Info
    //---------------------------------------
    } else if (packetId == 0x3cb) {
        long SOC = getUnsigned16(&bytes[2]);
        
        data.updateField(CANDataMgr::soc,                 SOC * 0.5);
        data.updateField(CANDataMgr::maxChargeCurrent,    bytes[1]);
        data.updateField(CANDataMgr::maxDischargeCurrent, -int(bytes[0]));
        data.updateField(CANDataMgr::minBattTemp,         convertTemp(bytes[5]));
        data.updateField(CANDataMgr::maxBattTemp,         convertTemp(bytes[4]));
        
    //---------------------------------------
    // 3CD - Voltage
    //---------------------------------------
    } else if (packetId == 0x3cd) {
        /// We're going to use only the voltage in packet 0x3B
        /// data.updateField(CANDataMgr::voltage, getUnsigned16(&bytes[2]));
    
    //---------------------------------------
    // 520 - Fuel Injector
    //---------------------------------------
    } else if (packetId == 0x520) {
        data.updateField(CANDataMgr::fuelInj, getUnsigned16(&bytes[1]));
        
    //---------------------------------------
    // 526 - VVTi
    //---------------------------------------
    } else if (packetId == 0x526) {
        data.updateField(CANDataMgr::VVTi, Round(bytes[1] / 2 - 16));					// Divided by 2 and subtracted 16 to convert to before/after TDC.
        
    //---------------------------------------
    // 529 - EV Mode
    //---------------------------------------
    } else if (packetId == 0x529) {
        std::string mode;
        
        switch (bytes[4]) {
            case 0x00: mode = "N";      break;
            case 0x40: mode = "EV";     break;
            case 0x80: mode = "Denied"; break;
            default:   mode = "Cancel"; break;
        }
        
        data.updateField(CANDataMgr::evMode, mode, true);
        
    //---------------------------------------
    // 52C - Coolant Temp
    //---------------------------------------
    } else if (packetId == 0x52c) {
        data.updateField(CANDataMgr::waterTemp, convertTemp(bytes[1] * 0.5));
        
    //---------------------------------------
    // 540 - Shifter
    //---------------------------------------
    } else if (packetId == 0x540) {

       /* enum ShiftPositionType 
             { UNKNOWN = -99,
               P = -2,
               R = -1,
               N =  0,
               D =  1,
               B =  2 };
	   */
		std::string shiftPosition;

        // ShiftPositionType shiftPosition = UNKNOWN;
        
        switch (bytes[1]) {
            case 0x00: shiftPosition = "B"; break;
            case 0x10: shiftPosition = "D"; break;
            case 0x20: shiftPosition = "N"; break;
            case 0x40: shiftPosition = "R"; break;
            case 0x80: shiftPosition = "P"; break;
        }
        
        // data.updateField(CANDataMgr::shiftLever, shiftPosition);
		data.updateField(CANDataMgr::shiftLever, shiftPosition, true);
        
    //---------------------------------------
    // 5A4 - Gas Gauge
    //---------------------------------------
    } else if (packetId == 0x5a4) {
        data.updateField(CANDataMgr::gasGauge, bytes[1]);
        
    //---------------------------------------
    // 5C8 - Cruise Control
    //---------------------------------------
    } else if (packetId == 0x5c8) {
        bool cruiseOn = bool(bytes[2] & 0x0010);  // test bit 4
        
        data.updateField(CANDataMgr::cruiseControl, cruiseOn ? 1 : 0);
        
    //---------------------------------------
    // 57F - Lights
    //---------------------------------------
    } else if (packetId == 0x57f) {
    
     /*   enum LightSwitchType 
             { UNKNOWN    = -99,
               Off        =   0,
               Parking    =   1,
               Headlights =   2,
               HighBeams  =   3 };
	 */
        
		std::string lightSwitch;

        // LightSwitchType lightSwitch = UNKNOWN;
        
        switch (bytes[1]) {
            case 0x00: lightSwitch = "Off";        break; 
            case 0x10: lightSwitch = "Parking";    break; 
            case 0x30: lightSwitch = "Headlights"; break; 
            case 0x38: lightSwitch = "HighBeams";  break; 
        }
        
        data.updateField(CANDataMgr::lights, lightSwitch, true);

	//---------------------------------------
    // 56D - Last calculations
    //---------------------------------------
	} else if (packetId == 0x56D) {

		// Calculate iKM/L or iMPG
		if ((speedNum <= 0) && (fuelFlowNum <= 0)){		// Set kmlMPG to 0 if speed and fuel flow are both 0
			kmlMPG = 0;
			data.updateField(CANDataMgr::kmlMPGnum, kmlMPG);
		}
			else if ((speedNum >= 0) && (fuelFlowNum <= 0)){		// Set kmlMPG to 999.99 if moving but fuel flow is 0
				kmlMPG = 999.9;
				data.updateField(CANDataMgr::kmlMPGnum, kmlMPG);
			}
					else {
						kmlMPG = Round(speedNum / fuelFlowNum, .1);		// Calculate iKM/L or iMPG and round to .1
						data.updateField(CANDataMgr::kmlMPGnum, kmlMPG);
					}

        
    //---------------------------------------
    // END
    //---------------------------------------
    }
}


//-------------------------------------------------------------------------------------
// isPacket returns packet id if the buffer points to a valid packet header
//-------------------------------------------------------------------------------------
PacketIdType CANPackets::isPacket(const BufferType* bufP) const {
    
    const PacketIdType  id  = getUnsigned16(bufP);
    const PacketLenType len = bufP[2];
    
    for (int i=0; i<NHEADERS; i++) {
        const HeaderDataType& head = headers[i];
        
        if (head.id  == id && 
            head.len == len)
        {
            return id;
        }
    }
    
    return PacketIdType(rFAILED);  // no packets found
}

